#include "localSearch.h"

#define LR_X 10
#define LR_Y 10
#define LR_F 0
#define LR_B 0

int LocalSearch::estimateSimilarPatchesStep1(std::vector<std::vector<float> > &o_group3d, std::vector<unsigned> &o_index, const unsigned pidx, bool rerank, bool excludeYourself, bool verbose)
{
	//! Initialization
	int sWx   = p_params.sizeSearchWindow;
	int sWy   = p_params.sizeSearchWindow;
	const int sWt_f = p_params.sizeSearchTimeRangeFwd;
	const int sWt_b = p_params.sizeSearchTimeRangeBwd;
	const int sPx   = p_params.sizePatch;
	const int sPt   = p_params.sizePatchTime;

	const VideoSize* i_im_sz = pm->infoVid();

	//! Coordinates of center of search box
	unsigned px, py, pt, pc;
	i_im_sz->coords(pidx, px, py, pt, pc);

	unsigned rangex[2];
	unsigned rangey[2];
	unsigned ranget[2];

#ifdef CENTRED_SEARCH
	rangex[0] = std::max(0, (int)px - (sWx-1)/2);
	rangey[0] = std::max(0, (int)py - (sWy-1)/2);
	ranget[0] = std::max(0, (int)pt -  sWt_b   );

	rangex[1] = std::min((int)i_im_sz->width  - sPx, (int)px + (sWx-1)/2);
	rangey[1] = std::min((int)i_im_sz->height - sPx, (int)py + (sWy-1)/2);
	ranget[1] = std::min((int)i_im_sz->frames - sPt, (int)pt +  sWt_f   );
#else
	int shift_x = std::min(0, (int)px - (sWx-1)/2); 
	int shift_y = std::min(0, (int)py - (sWy-1)/2); 
	int shift_t = std::min(0, (int)pt -  sWt_b   ); 

	shift_x += std::max(0, (int)px + (sWx-1)/2 - (int)i_im_sz->width  + sPx); 
	shift_y += std::max(0, (int)py + (sWy-1)/2 - (int)i_im_sz->height + sPx); 
	shift_t += std::max(0, (int)pt +  sWt_f    - (int)i_im_sz->frames + sPt); 

	rangex[0] = std::max(0, (int)px - (sWx-1)/2 - shift_x);
	rangey[0] = std::max(0, (int)py - (sWy-1)/2 - shift_y);
	ranget[0] = std::max(0, (int)pt -  sWt_b    - shift_t);

	rangex[1] = std::min((int)i_im_sz->width  - sPx, (int)px + (sWx-1)/2 - shift_x);
	rangey[1] = std::min((int)i_im_sz->height - sPx, (int)py + (sWy-1)/2 - shift_y);
	ranget[1] = std::min((int)i_im_sz->frames - sPt, (int)pt +  sWt_f    - shift_t);
#endif

	//! Redefine size of search range
	sWx = rangex[1] - rangex[0] + 1;
	sWy = rangey[1] - rangey[0] + 1;
	int sWt = ranget[1] - ranget[0] + 1;

	std::vector<std::pair<float, unsigned> > distance;

	//! Compute distance between patches in search range
	for (unsigned qt = ranget[0], dt = 0; qt <= ranget[1]; qt++, dt++)
	for (unsigned qy = rangey[0], dy = 0; qy <= rangey[1]; qy++, dy++)
	for (unsigned qx = rangex[0], dx = 0; qx <= rangex[1]; qx++, dx++)
	{
		if(excludeYourself && (qt == pt) && (qx > (px - sPx/2) && qx < (px + sPx/2) && qy > (py - sPx/2) && qy < (py + sPx/2)))
			continue;

		unsigned currentPatch = i_im_sz->index(qx, qy, qt, 0);
		distance.push_back(std::make_pair(pm->distanceOrigPos(pidx, currentPatch), currentPatch));
	}

	//! Keep only the N2 best similar patches
	const unsigned nSimP = std::min(kNN, (unsigned)distance.size());
	std::partial_sort(distance.begin(), distance.begin() + nSimP,
	                  distance.end(), comparaisonFirst);

	if (nSimP <  p_params.nSimilarPatches)
	{
		printf("SR1 [%d,%d,%d] ~ [%d-%d, %d-%d, %d-%d] - nsim = %d\n", 
				px,py,pt,rangex[0], rangex[1], rangey[0], rangey[1], ranget[0], ranget[1], nSimP);
	}

	//! Register position of patches
	for (unsigned n = 0; n < nSimP; n++)
	       	o_index[n] = distance[n].second;

	//! Stack selected patches into the 3D group
	pm->loadGroupStep1(o_group3d, o_index, kNN, pidx);

	return nSimP;
}

int LocalSearch::estimateSimilarPatchesStep2(Video<float> const& i_imNoisy, std::vector<float> &o_group3dNoisy, std::vector<float> &o_group3dBasic, std::vector<unsigned> &o_index, const unsigned pidx, bool rerank, bool excludeYourself, bool verbose)
{
	//! Initialization
	int sWx   = p_params.sizeSearchWindow;
	int sWy   = p_params.sizeSearchWindow;
	const int sWt_f = p_params.sizeSearchTimeRangeFwd;
	const int sWt_b = p_params.sizeSearchTimeRangeBwd;
	const int sPx   = p_params.sizePatch;
	const int sPt   = p_params.sizePatchTime;

	const VideoSize* i_im_sz = pm->infoVid();

	//! Coordinates of center of search box
	unsigned px, py, pt, pc;
	i_im_sz->coords(pidx, px, py, pt, pc);

	unsigned rangex[2];
	unsigned rangey[2];
	unsigned ranget[2];

#ifdef CENTRED_SEARCH
	rangex[0] = std::max(0, (int)px - (sWx-1)/2);
	rangey[0] = std::max(0, (int)py - (sWy-1)/2);
	ranget[0] = std::max(0, (int)pt -  sWt_b   );

	rangex[1] = std::min((int)i_im_sz->width  - sPx, (int)px + (sWx-1)/2);
	rangey[1] = std::min((int)i_im_sz->height - sPx, (int)py + (sWy-1)/2);
	ranget[1] = std::min((int)i_im_sz->frames - sPt, (int)pt +  sWt_f   );
#else
	int shift_x = std::min(0, (int)px - (sWx-1)/2); 
	int shift_y = std::min(0, (int)py - (sWy-1)/2); 
	int shift_t = std::min(0, (int)pt -  sWt_b   ); 

	shift_x += std::max(0, (int)px + (sWx-1)/2 - (int)i_im_sz->width  + sPx); 
	shift_y += std::max(0, (int)py + (sWy-1)/2 - (int)i_im_sz->height + sPx); 
	shift_t += std::max(0, (int)pt +  sWt_f    - (int)i_im_sz->frames + sPt); 

	rangex[0] = std::max(0, (int)px - (sWx-1)/2 - shift_x);
	rangey[0] = std::max(0, (int)py - (sWy-1)/2 - shift_y);
	ranget[0] = std::max(0, (int)pt -  sWt_b    - shift_t);

	rangex[1] = std::min((int)i_im_sz->width  - sPx, (int)px + (sWx-1)/2 - shift_x);
	rangey[1] = std::min((int)i_im_sz->height - sPx, (int)py + (sWy-1)/2 - shift_y);
	ranget[1] = std::min((int)i_im_sz->frames - sPt, (int)pt +  sWt_f    - shift_t);
#endif

	//! Redefine size of search range
	sWx = rangex[1] - rangex[0] + 1;
	sWy = rangey[1] - rangey[0] + 1;
	int sWt = ranget[1] - ranget[0] + 1;

	std::vector<std::pair<float, unsigned> > distance;

	//! Compute distance between patches in search range
	const int chnls = i_im_sz->channels;
	for (unsigned qt = ranget[0], dt = 0; qt <= ranget[1]; qt++, dt++)
	for (unsigned qy = rangey[0], dy = 0; qy <= rangey[1]; qy++, dy++)
	for (unsigned qx = rangex[0], dx = 0; qx <= rangex[1]; qx++, dx++)
	{
		if(excludeYourself && (qt == pt) && (qx > (px - sPx/2) && qx < (px + sPx/2) && qy > (py - sPx/2) && qy < (py + sPx/2)))
			continue;
		unsigned currentPatch = i_im_sz->index(qx, qy, qt, 0);
		//! Save distance and corresponding patch index
		distance.push_back(std::make_pair(pm->distanceOrigPos(pidx, currentPatch), currentPatch));
	}

	//! Keep only the nSimilarPatches best similar patches
	unsigned nSimP = std::min(kNN, (unsigned)distance.size());
	std::partial_sort(distance.begin(), distance.begin() + nSimP,
	                  distance.end(), comparaisonFirst);

	if (nSimP <  p_params.nSimilarPatches)
	{
		printf("SR2 [%d,%d,%d] ~ [%d-%d, %d-%d, %d-%d] - nsim = %d\n", 
				px,py,pt,rangex[0], rangex[1], rangey[0], rangey[1], ranget[0], ranget[1], nSimP);
	}

	//! Register position of similar patches
	for (unsigned n = 0; n < nSimP; n++)
			o_index[n] = distance[n].second;

	//! Save similar patches into 3D groups
	pm->loadGroupStep2(o_group3dBasic, o_index, i_imNoisy, o_group3dNoisy, nSimP, pidx);
	return nSimP;
}

int LocalSearch::estimateDistanceSimilarPatchesStep1(std::vector<float> &output, std::vector<unsigned> &index, const unsigned pidx, bool rerank, bool excludeYourself, bool verbose)
{
	//! Initialization
	int sWx   = p_params.sizeSearchWindow;
	int sWy   = p_params.sizeSearchWindow;
	const int sWt_f = p_params.sizeSearchTimeRangeFwd;
	const int sWt_b = p_params.sizeSearchTimeRangeBwd;
	const int sPx   = p_params.sizePatch;
	const int sPt   = p_params.sizePatchTime;

	const VideoSize* i_im_sz = pm->infoVid();

	//! Coordinates of center of search box
	unsigned px, py, pt, pc;
	i_im_sz->coords(pidx, px, py, pt, pc);

	unsigned rangex[2];
	unsigned rangey[2];
	unsigned ranget[2];

#ifdef CENTRED_SEARCH
	rangex[0] = std::max(0, (int)px - (sWx-1)/2);
	rangey[0] = std::max(0, (int)py - (sWy-1)/2);
	ranget[0] = std::max(0, (int)pt -  sWt_b);

	rangex[1] = std::min((int)i_im_sz->width  - sPx, (int)px + (sWx-1)/2);
	rangey[1] = std::min((int)i_im_sz->height - sPx, (int)py + (sWy-1)/2);
	ranget[1] = std::min((int)i_im_sz->frames - sPt, (int)pt +  sWt_f   );
#else
	int shift_x = std::min(0, (int)px - (sWx-1)/2); 
	int shift_y = std::min(0, (int)py - (sWy-1)/2); 
	int shift_t = std::min(0, (int)pt -  sWt_b   ); 

	shift_x += std::max(0, (int)px + (sWx-1)/2 - (int)i_im_sz->width  + sPx); 
	shift_y += std::max(0, (int)py + (sWy-1)/2 - (int)i_im_sz->height + sPx); 
	shift_t += std::max(0, (int)pt +  sWt_f    - (int)i_im_sz->frames + sPt); 

	rangex[0] = std::max(0, (int)px - (sWx-1)/2 - shift_x);
	rangey[0] = std::max(0, (int)py - (sWy-1)/2 - shift_y);
	ranget[0] = std::max(0, (int)pt -  sWt_b    - shift_t);

	rangex[1] = std::min((int)i_im_sz->width  - sPx, (int)px + (sWx-1)/2 - shift_x);
	rangey[1] = std::min((int)i_im_sz->height - sPx, (int)py + (sWy-1)/2 - shift_y);
	ranget[1] = std::min((int)i_im_sz->frames - sPt, (int)pt +  sWt_f    - shift_t);
#endif

	//! Redefine size of search range
	sWx = rangex[1] - rangex[0] + 1;
	sWy = rangey[1] - rangey[0] + 1;
	int sWt = ranget[1] - ranget[0] + 1;

	std::vector<std::pair<float, unsigned> > distance;

	//! Compute distance between patches in search range
	for (unsigned qt = ranget[0], dt = 0; qt <= ranget[1]; qt++, dt++)
		for (unsigned qy = rangey[0], dy = 0; qy <= rangey[1]; qy++, dy++)
			for (unsigned qx = rangex[0], dx = 0; qx <= rangex[1]; qx++, dx++)
			{
				if(excludeYourself && (qt == pt) && (qx > (px - sPx/2) && qx < (px + sPx/2) && qy > (py - sPx/2) && qy < (py + sPx/2)))
					continue;
				unsigned currentPatch = i_im_sz->index(qx, qy, qt, 0);

				//! Save distance and corresponding patch index
				distance.push_back(std::make_pair(pm->distanceOrigPos(pidx, currentPatch), currentPatch));
			}

	const unsigned nSimP = std::min(kNN, (unsigned)distance.size());
	std::partial_sort(distance.begin(), distance.begin() + nSimP,
			distance.end(), comparaisonFirst);

	if (nSimP <  p_params.nSimilarPatches)
	{
		printf("SR1 [%d,%d,%d] ~ [%d-%d, %d-%d, %d-%d] - nsim = %d\n", 
				px,py,pt,rangex[0], rangex[1], rangey[0], rangey[1], ranget[0], ranget[1], nSimP);
	}

	for (unsigned n = 0; n < nSimP; n++)
	{
		output[n] = distance[n].first;
		index[n] = distance[n].second;
	}

	return nSimP;
}
