#!/bin/bash
#set -x
# dctdenoising_multi.sh 
#        noisy.tif
#        40
#        out.tif      # also creates out.tif.non.tif
#        "-w 8 -1"    # dctdenoising params (optional)
#        3            # LEVELS of pyramid (default: -1 = auto) (optional)
#        2            # R_PYR pyramid ratio: 2 (optional)
#        0.7          # PAR_PYR recomposition ratio : 0.7 (optional)

export PATH=$PATH:/bin:/usr/local/bin:/usr/bin:./

if [ $# -lt 4 ]; then
   echo "$0 noisy.tif sigma out.tif"
   echo "    \"-w 8 -1\"  # dctdenoising params (optional)"
   echo "    -1           # LEVELS of pyramid (default: -1 = auto) (optional)"
   echo "    2            # R_PYR pyramid ratio: 2 (optional), 1.5 also possible"
   echo "    0.7          # PAR_PYR recomposition ratio : 0.7 (optional)"
   echo "    1 150        # first_frame and last_frame for the video"
   echo "    1            # 3D dct for the video instead of frame by frame"
   exit 1
fi

INPUT=$1
NOISE=$2
OUTPUT=$3
NLB_ARGS=$4
LEVELS=-1
FIRST=1
LAST=150
DCT3D=1
R_PYR=2   
PAR_PYR=0.7


if [ -n "$5" ]; then
   LEVELS=$5
fi
if [ -n "$6" ]; then
   FIRST=$6
fi
if [ -n "$7" ]; then
   LAST=$7
fi
if [ -n "$8" ]; then
   DCT3D=$8
fi
if [ -n "$9" ]; then
   R_PYR=$9
fi
if [ -n "$10" ]; then
   PAR_PYR=$10
fi

# determine the levels based on the image size
if [ $LEVELS -eq -1 ]; then
   PIXELS=$(./num_pixels $INPUT)
   echo $PIXELS
   if [ ${PIXELS} -lt 500000 ]; then
     LEVELS=1
   elif [ ${PIXELS} -lt 2000000 ]; then
     LEVELS=2
   elif [ ${PIXELS} -lt 8000000 ]; then
     LEVELS=3
   else
     LEVELS=4
   fi
   echo "Scales: $LEVELS (auto)"
else
   echo "Scales: $LEVELS (requested)"
fi


DEC_ARGS="-r ${R_PYR}"
MS_ARGS="-c ${PAR_PYR}"


#plambda ${INPUT} "x 0 *" -o level_zero.tiff

./addnoise -i ${INPUT} -sigma $NOISE -f ${FIRST} -l ${LAST}

./decompose_video ./noisy_%04d.png level_ ${LEVELS} _%04d.png ${FIRST} ${LAST} -3d ${DCT3D} ${DEC_ARGS}

lvl=$((LEVELS-1))
sigma=$(bc <<< "scale=2; $NOISE / ${R_PYR}^$lvl")
#./dctdenoising level_${lvl}.tiff ${sigma} level_${lvl}_non.tiff ${NLB_ARGS} 
./vnlbayes -i level_${lvl}_%04d.png -deno ./level_${lvl}_%04d_non.png -f ${FIRST} -l ${LAST} -px1 5 -pt1 4 -np1 375 -wx1 37 -wt1 2 -px2 5 -pt2 4 -np2 160 -wx2 37 -wt2 2 -tp 0 -sigma ${sigma} -r1 10000 -r2 37 -mngr 0 -rr 0 -nbT 1 -scale 1 -ratio 2


for ((lvl=LEVELS-2; lvl>=0; --lvl))
do
  sigma=$(bc <<< "scale=2; $NOISE / ${R_PYR}^$lvl")
#  ./dctdenoising level_${lvl}.tiff ${sigma} level_${lvl}_non.tiff ${NLB_ARGS}
./vnlbayes -i level_${lvl}_%04d.png -deno ./level_${lvl}_%04d_non.png -f ${FIRST} -l ${LAST} -px1 5 -pt1 4 -np1 375 -wx1 37 -wt1 2 -px2 5 -pt2 4 -np2 160 -wx2 37 -wt2 2 -tp 0 -sigma ${sigma} -r1 10000 -r2 37 -mngr 0 -rr 0 -nbT 1 -scale 1 -ratio 2
#  ./merge_coarse level_zero.tiff level_${lvl}_non.tiff level_${lvl}_non_Z.tiff ${MS_ARGS} &
done

wait
./recompose_video level_ ${LEVELS} _%04d.png $OUTPUT ${FIRST} ${LAST} -3d ${DCT3D} ${MS_ARGS}

#plambda level_0_non.tiff          "x" -o $OUTPUT.single.tif

