
/*
 * Original work: Copyright (c) 2016, Thibaud Ehret <ehret.thibaud@gmail.com>
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>

#include <string>
#include <sstream>
#include <sys/stat.h>

#include "Utilities/Utilities.h"
#include "Utilities/cmd_option.h"
#include "Utilities/nfa.h"

using namespace std;

/**
 * @file   main_detection.cpp
 * @brief  Main executable file for detection
 *
 * @author THIBAUD EHRET  <ehret.thibaud@gmail.com>
 **/


int main(int argc, char **argv)
{
	clo_usage("NL-Bayes image denoising using an external database");
	clo_help(" NOTE: Input (<) and output (>) sequences are specified by their paths in printf format.\n");

	//! Paths to input/output sequences
	using std::string;
	const string  input_path = clo_option("-i"    , ""              , "< input image (multiple images can be denoised at the same time using the C patch specification such as images%03d.tiff, parameter f and l are then used for the exact specification, all images must have the same size)");
	const string  mask_path = clo_option("-mask"    , ""              , "< input image (multiple images can be denoised at the same time using the C patch specification such as images%03d.tiff, parameter f and l are then used for the exact specification, all images must have the same size)");
	const string   diff_path = clo_option("-diff" , "diff_%03d.png" , "> difference sequence");
	const string   visual_path = clo_option("-vis" , "visual_%03d.png" , "> difference sequence");

	//! General parameters
	const int k1 = clo_option("-k1", 3, "< Add noise of standard deviation sigma");
	const int k2 = clo_option("-k2", 3, "< Add noise of standard deviation sigma");
	const long double nfa_threshold = clo_option("-nfath", 1., "< Add noise of standard deviation sigma");

	Video<float> original, diff, mask;
	original.loadVideo(input_path, 0, 0, 1);
	diff.loadVideo(diff_path, 0, 0, 1);


	if(mask_path != "")
	{
		mask.loadVideo(mask_path, 0, 0, 1);
		for(int x = 0; x < diff.sz.width; ++x)
		for(int y = 0; y < diff.sz.height; ++y)
		for(int c = 0; c < diff.sz.channels; ++c)
					diff(x,y,0,c) = mask(x,y,0,0) * diff(x,y,0,c);
	}

	Video<float> visual_nfa(original.sz.width, original.sz.height, original.sz.frames, 3, 0.);
	
	int type;
	long double nfa = computeNFAs(original, diff, k1, k2, visual_nfa, type, nfa_threshold);

	visual_nfa.saveVideo(visual_path, 1, 1, 1);

	{
		ofstream file;

		file.open("measures.txt", ios::out | ios::app);
		//! Check if the file is open
		if (!file) {
			return EXIT_FAILURE;
		}
		file << "Minimum logNFA = " << nfa << endl;
		if(type == 0)
			file << "Type = No problem detected"<< type << endl;
		else if(type == 1)
			file << "Type = At least one detection of type 1 and no detection of type 2" << endl;
		else if(type == 2)
			file << "Type = At least one detection of type 2 and no detection of type 1" << endl;
		else if(type == 3)
			file << "Type = At least one detection of type 1 and 2" << endl;
		file << "-----------------------" << endl;
		cout << endl;
		//! Close the file
		file.close();
	}


	return EXIT_SUCCESS;
}
