/*
 * Copyright (c) 2015, Pablo Arias <pariasm@gmail.com>
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>

#include <string>
#include <cstring>
#include <sstream>
#include <sys/stat.h>

#include "NlBayes/LibMatrix.h"
#include "Utilities/Utilities.h"
#include "NlBayes/VideoNLBayes.h"
#include "Utilities/cmd_option.h"

#include "Utilities/PatchSearch/localReinforcement.h"
#include "Utilities/PatchSearch/localPatchPropagation.h"

#include <algorithm>
#include "Utilities/PatchManager/videoPatchManager.h"
#include "Utilities/PatchManager/databasePatchManager.h"
#include "Utilities/PartitionTree/forestManager.h"
#include <ctime>

/**
 * @file   patch_group.cpp
 * @brief  Retrieves the group of similar of a given one.
 *
 *
 *
 * @author PABLO ARIAS  <pariasm@gmail.com>
 **/

float computeNRMSE(unsigned px, unsigned py, unsigned pt
		, unsigned qx, unsigned qy, unsigned qt
		, Video<float>& i_im, int sPx, int sPt)
{
	float dist = 0.f, dif;
	for (unsigned ht = 0; ht < sPt; ht++)
		for (unsigned hy = 0; hy < sPx; hy++)
			for (unsigned hx = 0; hx < sPx; hx++)
				for(unsigned hc = 0; hc < i_im.sz.channels; ++hc)
					dist += (dif = i_im(px + hx, py + hy, pt + ht, hc)
							- i_im(qx + hx, qy + hy, qt + ht, hc)) * dif;

	dist /= (float) (sPx * sPx * sPt * i_im.sz.channels);
	dist = std::sqrt(dist) / 255.f;	
	return dist;
}

void print_group_coordinates(
		std::vector<unsigned>& indexes 
		,	std::vector<float>& dists
		,	VideoSize sz
		,	std::string out
		){
	FILE *fid = out.size() ? fopen(out.c_str(),"a") : stdout;
	for (unsigned idx = 0; idx < indexes.size(); idx++) 
	{
		unsigned px, py, pt, pc;
		sz.coords(indexes[idx], px, py, pt, pc);
		fprintf(fid, "[COORD]% 3d % 3d % 3d\n", px, py, pt);
		fprintf(fid, "[DISTA]%f\n", dists[idx]);
	}
	if (out.size()) fclose(fid);

}

void print_spectrum(
		std::vector<float>& eigenvalues 
		,	std::vector<float>& projection
		,	std::string out
		){
	FILE *fid = out.size() ? fopen(out.c_str(),"a") : stdout;
	for (unsigned idx = 0; idx < eigenvalues.size(); idx++) 
	{
		fprintf(fid, "[EIG] %f  %f\n", std::sqrt(std::abs(eigenvalues[eigenvalues.size() - idx - 1])), std::abs(projection[eigenvalues.size() - idx - 1]));
	}
	if (out.size()) fclose(fid);
}

void export_spectrum(
		Video<float>& visual
		, std::vector<float>& baricenter 
		, std::vector<float>& eigenvectors 
		, unsigned sPx 
		, unsigned sPC 
		, unsigned channels
		, unsigned nSimP
		){
	const unsigned taille_bande = 3;
	const unsigned nb_lignes = 2;
	visual.resize(nSimP * sPx + nSimP*taille_bande, nb_lignes * sPx + (nb_lignes-1)*taille_bande, 1, channels);
	unsigned x,y;
	y = 0;
	x = 0;
	for(unsigned pidx = 0; pidx < 1; ++pidx)
	{
		for(int c = 0, k = 0; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i,y + j, 0, c) = baricenter[k];	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
	for(unsigned pidx = 1; pidx < nSimP; ++pidx)
	{
		for(int c = 0, k = 0; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i,y + j, 0, c) = 0;	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
	y += sPx;
	for(int i = 0; i < visual.sz.width; ++i)
	{
		for(int j = 0; j < taille_bande; ++j)
		{
			visual(i, y + j, 0, 0) = 255.;
		}
	}
	y += taille_bande;
	x = 0;
	for(unsigned pidx = 0; pidx < nSimP; ++pidx)
	{
		for(int c = 0, k = (nSimP- pidx - 1)*sPC; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i, y + j, 0, c) = std::max(std::min(255*(4*eigenvectors[k] + 0.5), 255.), 0.);	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
}

//void print_group(
//		std::vector<float> &group3d
//		,	const unsigned patch_dim
//		,	const unsigned patch_num
//		,	VideoSize sz
//		,	std::string prefix
//		,	std::string out
//		){
//	FILE *fid = out.size() ? fopen(out.c_str(),"a") : stdout;
//	for (unsigned idx = 0; idx < patch_num; idx++) 
//	{
//		fprintf(fid, "%s", prefix.c_str());
//		for (unsigned k = 0; k < patch_dim; k++)
//			fprintf(fid, "% 3f   ", group3d[k*patch_num + idx]);
//		fprintf(fid, "\n");
//	}
//	if (out.size()) fclose(fid);
//}

void export_groups(Video<float>& visual, std::vector<float> patch, std::vector<float> patchgroup1, std::vector<float> patchgroup2, unsigned nSimP, unsigned sPx, unsigned channels)
{
	const unsigned nb_lignes = 3;
	const unsigned taille_bande = 3;
	visual.resize(nSimP * sPx + nSimP*taille_bande, nb_lignes * sPx + (nb_lignes-1)*taille_bande, 1, channels);

	for(unsigned i = 0; i < visual.sz.width; ++i)
	for(unsigned j = 0; j < visual.sz.height; ++j)
	for(unsigned c = 0; c < channels; ++c)
		visual(i,j,0,c) = 0;

	// load patch
	unsigned x,y,n;
	y = 0;
	x = 0;
	for(unsigned pidx = 0; pidx < 1; ++pidx)
	{
		for(int c = 0, k = 0; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i,y + j, 0, c) = patch[k];	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
	for(unsigned pidx = 1; pidx < nSimP; ++pidx)
	{
		for(int c = 0, k = 0; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i,y + j, 0, c) = 0;	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
	y += sPx;

	for(int i = 0; i < visual.sz.width; ++i)
	{
		for(int j = 0; j < taille_bande; ++j)
		{
			visual(i, y + j, 0, 0) = 255.;
		}
	}
	y += taille_bande;

	// load patchgroup1
	x = 0;
	for(unsigned pidx = 0; pidx < nSimP; ++pidx)
	{
		for(int c = 0, k = 0; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i,y + j, 0, c) = patchgroup1[k * nSimP + pidx];	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
	y += sPx;
	for(int i = 0; i < visual.sz.width; ++i)
	{
		for(int j = 0; j < taille_bande; ++j)
		{
			visual(i, y + j, 0, 0) = 255.;
		}
	}
	y += taille_bande;

	// load patchgroup2
	x = 0;
	for(unsigned pidx = 0; pidx < nSimP; ++pidx)
	{
		for(int c = 0, k = 0; c < channels; ++c)
		for(int j = 0; j < sPx; ++j)
		for(int i = 0; i < sPx; ++i, ++k)
		{
			visual(x + i,y + j, 0, c) = patchgroup2[k * nSimP + pidx];	
		}
		x += sPx;

		// rempli une bande
		for(int i = 0; i < taille_bande; ++i)
		{
			for(int j = 0; j < sPx; ++j)
			{
				visual(x + i, y + j, 0, 0) = 255.;
			}
		}
		x += taille_bande;
	}
}

int main(int argc, char **argv)
{
	clo_usage("Retrieves group of similar patches and outputs it in ASCII files");

	//! Get command line inputs
	const string input_path = clo_option("-i", "", "Path to input sequence (printf format)");
	const string data_path = clo_option("-data", "", "Path to input sequence (printf format)");

	const string output_path  = clo_option("-out", "stdout", "Output path");

	const int i_firstFrame = clo_option("-f", 0, "First frame");
	const int i_lastFrame  = clo_option("-l", 0, "Last frame");
	const int i_frameStep  = clo_option("-s", 1, "Frame step");

	const unsigned px      = clo_option("-PAx", 0, "Point of analysis");
	const unsigned py      = clo_option("-PAy", 0, "Point of analysis");
	const unsigned pt      = clo_option("-PAt", 0, "Point of analysis");

	//! Video NLB parameters relevant to the computation of distances
	const int patch_sizex  = clo_option("-px",-1, "> Spatial patch size, step 1");
	const int patch_sizet  = clo_option("-pt", 1, "> Temporal patch size, step 1");
	const unsigned patch_num  = clo_option("-np",-1, "> Number of similar patches, step 1");

	const unsigned nSimP = patch_num;
	//! Load video
	Video<float> input, data;
	input.loadVideo(input_path.c_str(), i_firstFrame, i_lastFrame, i_frameStep);
	data.loadVideo(data_path.c_str(), i_firstFrame, i_lastFrame, i_frameStep);

	//! Set parameters
	VideoNLB::nlbParams prms;
	{
		//! Initialize parameter structures
		VideoNLB::initializeNlbParameters(prms, i_frameStep, 0, input.sz, 0, 1, 0, 0, patch_sizet);

		//! Override with command line parameters
		if (patch_sizex  >= 0) VideoNLB::setSizePatch(prms, input.sz, (unsigned)patch_sizex);;
		if (patch_num  >= 0) VideoNLB::setNSimilarPatches(prms, (unsigned)patch_num);

		//! Print parameters
		VideoNLB::printNlbParameters(prms);
	}

	prms.nbTrees = 4;

	PatchManager* dbpm;
	dbpm = new DatabasePatchManager(data, patch_sizex, patch_sizet, data.sz.channels);
	PatchManager* pm;
	pm = new VideoPatchManager(data, patch_sizex, patch_sizet, data.sz.channels);
	std::vector<PartitionTree*> trees; 
	//if(prms2.type > 0)
	{
		PrmTree paramTree;
		paramTree.epsilon = 0.; 
		paramTree.prms = prms;
		paramTree.nbVP = 1;
		paramTree.subS = 1000;
		paramTree.rand = 5;
#ifdef _OPENMP
#pragma omp parallel for schedule(dynamic) \
		shared(trees) 
#endif
		for(int i = 0; i < prms.nbTrees; ++i)
		{
			char filename[128];
			sprintf(filename, "vp%d.save", i);
			trees.push_back(new VPTree(*pm, paramTree, filename));	
		}
	}
	ForestManager dbfm(*pm, trees, prms.nSimilarPatches, prms);
	dbpm->setCurrentVideo(input);
	dbfm.updatePM(*dbpm);
	ForestManager fm(*pm, trees, prms.nSimilarPatches, prms);

	LocalPatchPropagation* lpp;
	lpp = new LocalReinforcement(*dbpm, dbfm, patch_num, prms);
	LocalPatchPropagation* lpp2;
	lpp2 = new LocalReinforcement(*pm, fm, patch_num, prms);

	unsigned sPC = patch_sizex * patch_sizex * patch_sizet * input.sz.channels;
	unsigned sPx = patch_sizex;
	unsigned patch_dim = sPC;
	unsigned channels = input.sz.channels;

	// Current patch
	unsigned pidx = input.sz.index(px, py, pt, 0);

	std::vector<unsigned> unindex(1, pidx);
	std::vector<float> patch1(sPC);
	std::vector<float> patch2(sPC);
	dbpm->loadGroup(patch1, unindex, 1, pidx);
	dbpm->loadGroup(patch2, unindex, 1, pidx);

	std::vector<float> distances1(nSimP);
	std::vector<float> distances2(nSimP);
	std::vector<unsigned> index1(nSimP);
	std::vector<unsigned> index2(nSimP);

	std::vector<float> patchgroup1(nSimP * sPC);
	std::vector<float> patchgroup2(nSimP * sPC);
	std::vector<float> groupEmpty(nSimP * sPC);

	matWorkspace i_mat1;
	i_mat1.groupTranspose.resize(patch_num * patch_dim);
	i_mat1.tmpMat          .resize(patch_dim * patch_dim);
	i_mat1.covMat          .resize(patch_dim * patch_dim);
	i_mat1.covMatTmp       .resize(patch_dim * patch_dim);
	i_mat1.baricenter      .resize(patch_dim);

	matWorkspace i_mat2;
	i_mat2.groupTranspose.resize(patch_num * patch_dim);
	i_mat2.tmpMat          .resize(patch_dim * patch_dim);
	i_mat2.covMat          .resize(patch_dim * patch_dim);
	i_mat2.covMatTmp       .resize(patch_dim * patch_dim);
	i_mat2.baricenter      .resize(patch_dim);
		
	lpp->estimateDistanceSimilarPatchesStep1(distances1, index1, pidx, false, true);
	lpp2->estimateDistanceSimilarPatchesStep1(distances2, index2, index1[0], false, false);

	pm->loadGroupStep2(patchgroup1, index1, input, groupEmpty, nSimP, pidx);
	pm->loadGroupStep2(patchgroup2, index2, input, groupEmpty, nSimP, pidx);

	Video<float> visual;
	export_groups(visual, patch1, patchgroup1, patchgroup2, nSimP, sPx, channels);
	unsigned r = sPC;
	// CAS 1
	centerData(patchgroup1, i_mat1.baricenter, nSimP, sPC);
	centerDataWith(patch1, i_mat1.baricenter, 1, sPC);
	covarianceMatrix(patchgroup1, i_mat1.covMat, nSimP, sPC);
	int info = matrixEigs(i_mat1.covMat, sPC, r, i_mat1.covEigVals, i_mat1.covEigVecs);
	productMatrix(i_mat1.groupTranspose,
			patch1,
			i_mat1.covEigVecs,
			1, r, sPC,
			false, false);

	// CAS 2
	centerData(patchgroup2, i_mat2.baricenter, nSimP, sPC);
	centerDataWith(patch2, i_mat2.baricenter, 1, sPC);
	covarianceMatrix(patchgroup2, i_mat2.covMat, nSimP, sPC);
	info = matrixEigs(i_mat2.covMat, sPC, r, i_mat2.covEigVals, i_mat2.covEigVecs);
	productMatrix(i_mat2.groupTranspose,
			patch2,
			i_mat2.covEigVecs,
			1, r, sPC,
			false, false);

	// Now that we have computed everything that we need, we export the rest
	char name[124];
	sprintf(name, "group1.out");
	char output_file_local[512];
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	print_group_coordinates(index1, distances1, input.sz, output_file_local);

	sprintf(name, "group2.out");
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	print_group_coordinates(index2, distances2, input.sz, output_file_local);

	sprintf(name, "spectrum1.out");
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	print_spectrum(i_mat1.covEigVals, i_mat1.groupTranspose, output_file_local);

	Video<float> pca1;
	sprintf(name, "pca1.tiff");
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	export_spectrum(pca1, i_mat1.baricenter, i_mat1.covEigVecs, sPx, sPC, channels, sPC);
	pca1.saveVideo(output_file_local, 1, 1);

	Video<float> pca2;
	sprintf(name, "pca2.tiff");
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	export_spectrum(pca2, i_mat2.baricenter, i_mat2.covEigVecs, sPx, sPC, channels, sPC);
	pca2.saveVideo(output_file_local, 1, 1);

	sprintf(name, "spectrum2.out");
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	print_spectrum(i_mat2.covEigVals, i_mat2.groupTranspose, output_file_local);

	sprintf(name, "visual.tiff");
	strcpy(output_file_local, output_path.c_str());
	strcat(output_file_local, name); 
	visual.saveVideo(output_file_local, 1, 1);

	return 0;
}
